from kivy.uix.anchorlayout import AnchorLayout
from kivy.uix.label import Label
from kivy.graphics import Fbo, Color, Rectangle, Scale
from kivy.properties import ColorProperty, StringProperty, ObjectProperty
from kivy.utils import platform
from threading import Thread, Event


if platform == 'android':
    from .preview_camerax import PreviewCameraX as CameraPreview
else:
    from .preview_kivycamera import PreviewKivyCamera as CameraPreview
    from .preview_kivycamera import KivyCameraProviderInfo
    
class CameraProviderInfo():
    def get_name(self):
        if platform == 'android':
            provider = 'android'
        else:
            provider = KivyCameraProviderInfo().get_name()
        return provider

class Preview(AnchorLayout):

    ##########################################
    # Layout Properties
    ##########################################

    aspect_ratio      = StringProperty()
    orientation       = StringProperty()
    letterbox_color   = ColorProperty('black')
    filepath_callback = ObjectProperty()
    inhibit_property = False
    preview = None

    ##########################################
    # Camera Events
    ##########################################

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.anchor_x = 'center'
        self.anchor_y = 'center'
        self.label = Label()
        self.preview = CameraPreview()
        self.add_widget(self.label)
        self.add_widget(self.preview)
        self.inhibit_property = False
        for key in ['letterbox_color', 'aspect_ratio',
                    'orientation']:
            if key in kwargs:
                setattr(self, key, kwargs[key])
                if key == 'aspect_ratio':
                    self.preview.set_aspect_ratio(kwargs[key])
                if key == 'orientation':
                    self.preview.set_orientation(kwargs[key])
        self._fbo = None
        self._busy = False
        self.camera_connected = False
        self._image_available = Event()
        self.analyze_resolution = 1024
        self.auto_analyze_resolution = []
    
    def on_orientation(self,instance,orientation):
        if self.preview and not self.inhibit_property:
            self.preview.set_orientation(orientation)

    def on_aspect_ratio(self,instance, aspect_ratio):
        if  self.preview and not self.inhibit_property:
            self.preview.set_aspect_ratio(aspect_ratio)

    def on_size(self, layout, size):
        self.label.canvas.clear()
        with self.label.canvas:
            Color (*self.letterbox_color) 
            Rectangle(pos = self.pos, size = self.size)

    ##########################################
    # User Events - All Platforms
    ##########################################

    def connect_camera(self, analyze_pixels_resolution = 1024,
                       enable_analyze_pixels = False, **kwargs):
        self.analyze_resolution = analyze_pixels_resolution
        self.inhibit_property = True
        self.camera_connected = True
        self._fbo = None
        if enable_analyze_pixels:
            Thread(target=self.image_scheduler, daemon=True).start()
        self.preview.connect_camera(analyze_callback =
                                        self.analyze_image_callback_schedule,
                                    analyze_proxy_callback =
                                        self.analyze_imageproxy_callback,
                                    canvas_callback =
                                        self.possible_canvas_callback,
                                    **kwargs)

    def disconnect_camera(self):
        self._image_available.set()
        self.camera_connected = False
        self.preview.disconnect_camera()
        self.inhibit_property = False

    def capture_screenshot(self, **kwargs):
        self.preview.capture_screenshot(**kwargs)

    def select_camera(self, camera_id):
        return self.preview.select_camera(camera_id)

    ##########################################
    # User Events - some platforms
    ##########################################

    def capture_photo(self, **kwargs):
        self.preview.capture_photo(**kwargs)

    def capture_video(self, **kwargs):
        self.preview.capture_video(**kwargs)

    def stop_capture_video(self):
        self.preview.stop_capture_video()

    ##########################################
    # User Events - Android Only
    ##########################################

    def flash(self, state = None):
        return self.preview.flash(state)

    def torch(self, state):
        return self.preview.torch(state)

    def focus(self, x, y):
        self.preview.focus(x, y)

    def zoom(self, delta_scale):
        self.preview.zoom(delta_scale)

    ##########################################
    # Data Analysis, Image Size and Schedule
    ##########################################

    def analyze_image_callback_schedule(self, texture, tpos, tscale, mirror):
        # texture : Kivy Texture with same orientation as the Preview
        # tpos   : location of texture in Preview
        # tscale : scale from oriented Texture resolution to Preview resolution
        # mirror : true if preview is mirrored
        if not self._busy:
            self._busy = True
            # Create a texture with lower resolution
            if self.auto_analyze_resolution:
                # resolution set by the analyzer [w,h] regardless of
                # Preview orientation or aspect ratio.
                # If the aspect ratio is not the same the Fbo is distorted.
                # self.scale is a two element array
                fbo_size = self.auto_analyze_resolution
                scale = [tscale * texture.width / fbo_size[0],
                         tscale * texture.height / fbo_size[1]]
            else:
                # resolution is 'self.analyze_resolution' along the long edge
                # default value is 1024
                # Optionally set as a connect option.
                # Value is never greater that the sensor resolution.
                # The aspect ratio is always the same as the Preview
                # self.scale is a scalar
                fbo_scale = max(max(texture.size) / self.analyze_resolution, 1)
                fbo_size  = (round(texture.size[0]/fbo_scale),
                             round(texture.size[1]/fbo_scale))
                scale = tscale * fbo_scale
            origin    = (round(fbo_size[0]/2), round(fbo_size[1]/2))
            # new or resized texture
            if not self._fbo or self._fbo.size[0] != fbo_size[0] or\
               self._fbo.size[1] != fbo_size[1]:
                self._fbo = Fbo(size = fbo_size)
            self._fbo.clear()
            with self._fbo:
                Color(1,1,1,1)
                Scale(1,-1,1, origin = origin)
                Rectangle(texture= texture, size = fbo_size)
            self._fbo.draw()

            # save these for self.analyze_pixels_callback()
            self.pixels = self._fbo.texture.pixels
            self.im_size = self._fbo.texture.size
            self.scale = scale  # 2 ele list , or scalar
            self.tpos = tpos
            self.mirror = mirror
            # ready
            self._image_available.set()

    def image_scheduler(self):
        while True:
            self._image_available.wait()
            self._image_available.clear()
            if not self.camera_connected:
                break
            # Must pass pixels not Texture, becuase we are in a different
            # Thread
            self.analyze_pixels_callback(self.pixels, self.im_size, self.tpos,
                                         self.scale, self.mirror)
            self._busy = False

    def possible_canvas_callback(self, texture, tex_size, tex_pos):
        if self.camera_connected:
            self.canvas_instructions_callback(texture, tex_size, tex_pos)

    ##########################################
    # Data Analysis Callbacks 
    ##########################################
    
    # analyze_pixels_callback()
    #
    # pixels        : Kivy Texture pixels, always RGBA
    # image_size    : size of pixels
    # image_pos     : Bottom left corner of analysis Texture inside the
    #    Preview. AKA the letterbox size plus modified aspect ratio adjustment.
    # image_scale   : Ratio between the analyzed Texture resolution and
    #    screen image resolution.
    # mirror        : True if Preview is mirrored
    
    def analyze_pixels_callback(self, pixels, image_size, image_pos,
                                image_scale, mirror):
        pass

    # canvas_instructions_callback()
    #
    # texture  : the default texture to be displayed in the Priview
    # tex_size : texture size with mirror information
    # tex_pos  : texture pos with mirror information
    def canvas_instructions_callback(self, texture, tex_size, tex_pos):
        pass

    # analyze_imageproxy_callback()
    # Android only
    #
    # image_proxy :
    #   https://developer.android.com/reference/androidx/camera/core/ImageProxy
    # image_pos   : Bottom left corner of screen image insie the Preview, the
    #    letterbox size.
    # image_scale : Scale image_proxy size to screen image size.
    # mirror      : True if Preview is mirrored
    # degrees     : clockwise rotation required to make image_proxy the same
    #    orientation as the screen.
    def analyze_imageproxy_callback(self, image_proxy, image_pos, image_scale,
                                    mirror, degrees):
        pass





from kivy.app import App
from kivy.core.window import Window
from threading import Thread
from kivy.clock import mainthread
from kivy.utils import platform
from kivy.core import core_select_lib
from kivy.graphics import Rectangle, Color
from kivy.graphics.texture import Texture
from kivy.core.text import Label as CoreLabel
from kivy.metrics import sp
from gestures4kivy import CommonGestures
from camera4kivy.preview_common import PreviewCommon
if platform in ['macosx', 'ios']:
    from kivy.core.camera import Camera
else:
    from camera4kivy.based_on_kivy_core.camera import Camera

from kivy.logger import Logger

class KivyCameraProviderInfo():
    def get_name(self):
        if Camera:
            provider = str(Camera).split('.')[-2].split('_')[-1] 
        else:
            provider = ""
        return provider


class PreviewKivyCamera(PreviewCommon, CommonGestures):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.error_message = ''
        self.mirror = True
        self.switching_camera = False
        self.starting_camera = False
        self.abort_camera_start = False
        self.enable_zoom_gesture = False
        self.enable_focus_gesture  = False
        self.audio  = False
        self.cg_zoom_level = [1 , 1]
        self.window_width = Window.width
        if platform == 'ios':
            self._enable_on_resume()
        self.provider = KivyCameraProviderInfo().get_name()
            
    def __del__(self):
        self.disconnect_camera()

    @mainthread
    def _enable_on_resume(self):
        app = App.get_running_app()
        app.bind(on_resume = self.on_resume)
            
    def on_resume(self, arg):
        Window.update_viewport()

    def on_size(self, instance, size):
        self.configure_viewport()
        self.configure_texture_crop(None)
        if platform == 'ios' and self.window_width != Window.width:
            if self._camera:
                orientation = self._camera.get_device_orientation()
                if orientation in [1,2,3,4]:
                    self._camera.set_video_orientation(orientation)
        self.canvas.clear()
        if self.error_message:
            self.canvas_text(self.error_message)
        elif self._camera and self._camera._texture:
            self.on_tex(None)
        else:
            with self.canvas:
                Color(1,1,1,1)
                Rectangle(size = self.view_size, pos = self.view_pos)
        self.window_width = Window.width

    #############################################
    # User Events
    #############################################

    def connect_camera(self,
                       camera_id = '0',
                       mirrored = True,
                       audio = False,
                       sensor_resolution = [],
                       sensor_rotation = 0,
                       default_zoom = 1.0,
                       enable_zoom_gesture = True,
                       enable_focus_gesture = True,                         
                       filepath_callback = None,
                       analyze_callback = None,
                       canvas_callback = None,
                       **kwargs):
        self.set_index(camera_id)
        if audio == True:
            self.audio = True
        self.set_resolution(sensor_resolution)
        self.set_rotation(sensor_rotation)
        self.set_filepath_callback(filepath_callback)
        self.data_callback = analyze_callback
        self.canvas_callback = canvas_callback
        self.default_zoom = min(max(default_zoom,0),1)
        self.enable_zoom_gesture = enable_zoom_gesture
        self.enable_focus_gesture = enable_focus_gesture        
        self.cg_zoom_level = [self.default_zoom, self.default_zoom]
        if platform == 'ios':
            self.mirror = self.index != 0
        else:
            self.mirror = mirrored
        self.stop_camera()
        #Thread(target=self.start_camera, daemon=True).start()
        self.start_camera()

    def disconnect_camera(self):
        if self.starting_camera:
            # This was related to the Thread above, now redundant?
            self.abort_camera_start = True
        else:
            self.stop_camera()
        
    def select_camera(self, index):
        if self.switching_camera or self.starting_camera:
            return self.index
        self.switching_camera = True
        if platform == 'ios':
            if self._camera:
                self.set_index(index)
                self.mirror = self.index != 0
                self._camera.change_camera_input(self.index)
                self.zoom_abs(self.cg_zoom_level[self.index])                
        else:
            self.stop_camera()
            self.set_index(index)
            self.start_camera()
        self.switching_camera = False
        return index

    # Screenshot
    ######################
    def capture_screenshot(self, location = '.', subdir = '', name = ''):
        view_crop = self.screenshot_crop()
        path = self.capture_path(location, subdir, name, '.jpg')
        tex = self.export_as_image().texture.get_region(*view_crop)
        tex.flip_vertical()
        if platform == 'ios':
            self._camera.save_texture(tex, path)
        else:
            tex.save(path, flipped = False)   
        if self.callback:
            self.callback(path)

    # Photo
    ######################
    def capture_photo(self, location = '.', subdir = '', name = ''):
        if self._camera and self._camera.texture:
            path = self.capture_path(location, subdir, name, '.jpg')
            tex = self._camera.texture.get_region(*self.tex_crop)
            if platform == 'ios':
                self._camera.save_texture(tex, path)
            elif self.provider in ['picamera2', 'opencv']:
                self._camera.photo(path, self.callback)
                return
            else:
                tex.save(path, flipped = False)    
            if self.callback:
                self.callback(path)

    # Video
    ######################

    def capture_video(self, location = '', subdir = '', name = ''):
        if self._camera and self._camera.texture:
            if self.provider in ['picamera2', 'opencv']:
                path = self.capture_path(location, subdir, name, '.mp4')
                self._camera.video_start(path, self.callback)
            
    def stop_capture_video(self):
        if self._camera and self._camera.texture:
            if self.provider in ['picamera2', 'opencv']:
                self._camera.video_stop()
    
    ##############################
    # Preview Widget Touch Events
    ##############################

    # pinch/spread for zoom
    def cgb_zoom(self, touch0, touch1, x, y, scale):
        if self._camera and self.enable_zoom_gesture:
            if platform == 'ios':
                level = max(self.cg_zoom_level[self.index] * scale, 1)
                self.cg_zoom_level[self.index] = level 
                self.zoom_abs(level)
            elif self.provider in ['picamera2']:
                self._camera.zoom(scale)   

    # drag
    def cgb_drag(self, touch, x, y, dx, dy):
        if self._camera and self.enable_zoom_gesture:
            if self.provider in ['picamera2']:
                # normalize to preview image
                crop = self.screenshot_crop()
                dx = dx / crop[2]
                dy = dy / crop[3]
                self._camera.drag(dx, dy)
        
    #############################################
    # iOS only User Events
    #############################################

    def zoom_abs(self, level):
        if platform == 'ios' and self._camera:
            self._camera.zoom_level(level)

    #############################################
    # Picamera2 only User Events
    #############################################

    def zoom_delta(self, delta_scale):
        if self._camera and self.provider in ['picamera2']:
            self._camera.zoom(delta_scale)

    def drag(self, delta_x, delta_y):
        if self._camera and self.provider in ['picamera2']:
            crop = self.screenshot_crop()
            dx = delta_x / crop[2]
            dy = delta_y / crop[3]
            self._camera.drag(dx, dy)

    #############################################
    # Ignored User Events
    #############################################

    def flash(self, state):
        return 'off'

    def torch(self, state):
        return 'off'

    def focus(self, x, y):
        pass

    #############################################
    # Parse Arguments
    #############################################

    def set_index(self, index):
        index = index.lower()
        try:
            int(index)
            isint = True
        except:
            isint = False
        if isint:
            self.index = int(index)
        elif index == 'front':
            self.index = 1
        elif index == 'back':
            self.index = 0
        elif index == 'toggle' and self.index == 0:
            self.index = 1
        elif index == 'toggle' and self.index == 1:
            self.index = 0
        else:
            self.index = 0;     

    #############################################
    # Camera Events
    #############################################

    def start_camera(self):
        self.starting_camera = True
        try:
            if not self._sensor_resolution:
                # a max resolution will fall back to the highest available
                # except picamera
                if platform in ['macosx', 'ios']:
                    # default 16:9
                    self._sensor_resolution = [3840, 2160]
                elif self.provider in ['picamera','opencv']:
                    self._sensor_resolution = [1280 , 960]
                elif self.provider in ['picamera2']:
                    self._sensor_resolution = [800 , 600]
                else:
                    #default 4:3 , value ignored by gi
                    self._sensor_resolution = [6400, 4800]

            if self.provider in ['picamera2', 'opencv']:
                context = self
            else:
                context = None
            self._camera = Camera(index= self.index,
                                  resolution = self._sensor_resolution,
                                  rotation = self._sensor_rotation,
                                  callback = self.camera_error,
                                  context = context)
            self.error_message = ""
        except AttributeError as e:
            #Logger.warning(str(e))
            self.camera_error_message()
        except Exception as e:
            #Logger.warning(str(e))
            if self._camera:
                self.error_message = 'ERROR: Camera internal error.'
            else:
                self.error_message = 'ERROR: No camera provider found.'
            self._camera = None

        if self.error_message:
            self.canvas_text(self.error_message)
            
        if self._camera:
            self._camera.bind(on_load=self.configure_texture_crop)
            self._camera.bind(on_texture=self.on_tex)
            self._camera.start()
            self.zoom_delta(self.default_zoom)
        if self.abort_camera_start:
            self.stop_camera()
            self._camera = None
        self.abort_camera_start = False
        self.starting_camera = False

    def on_load(self):
        self.configure_texture_crop(None)

    def on_texture(self):
        self.on_tex(None)

    def camera_error(self):
        self.camera_error_message()
        self.canvas_text(self.error_message)            

    def camera_error_message(self):
        self.error_message = "WARNING: Unable to connect to camera_id '" +\
           str(self.index)+"'.\n" +\
           'Check that the camera is connected.'
        self._camera = None

    def stop_camera(self):
        if self._camera:
            self._camera.stop()
            self._camera.unbind(on_texture=self.on_tex)
            self.clear_texture()
            if self._camera.__class__.__name__ == 'CameraGi':
                self._camera.unload()
            del self._camera
            self._camera = None
            
    #############################################
    # Texture 
    #############################################

    def clear_texture(self):
        if self._camera and self._camera.texture:
            tex_size = self._camera.texture.size
            buf = bytes([255] * tex_size[0] * tex_size[1] * 3)
            fmt = self._camera._format   ## all the providers are 3 byte
            self._camera.texture.blit_buffer(buf, colorfmt= fmt,
                                             bufferfmt='ubyte')
            self.on_tex(None)
            

    def on_tex(self, camera):
        if self._camera and self._camera.texture:
            tex = self._camera.texture.get_region(*self.tex_crop)

            if self.data_callback:
                self.data_callback(tex, self.view_pos,
                                   self.tscale, self.mirror)
            if self.mirror:
                view_size = (-self.view_size[0], self.view_size[1])
                view_pos = (self.view_pos[0] + self.view_size[0],
                            self.view_pos[1])
            else:
                view_size = self.view_size
                view_pos  = self.view_pos
            self.canvas.clear()
            with self.canvas:
                Color(1,1,1,1)
                Rectangle(texture= tex, size = view_size, pos = view_pos)
                if self.canvas_callback:
                    self.canvas_callback(tex, view_size, view_pos)

    def configure_texture_crop(self, dontcare):
        if not self._camera or not self._camera.texture:
            return
        width_tex, height_tex = self._camera.texture.size
        self.tex_crop = self.crop_for_aspect_orientation(width_tex, height_tex)
        self.tscale = self.view_size[1] / self.tex_crop[3]

    def crop_for_aspect_orientation(self, width_tex, height_tex):
        orientation = self.decode_orientation()
        if self.aspect_ratio == "4:3":
            aspect = 4 / 3
        else:
            aspect = 16 / 9        
        crop_pos_x = 0
        crop_pos_y = 0
        crop_siz_x = width_tex
        crop_siz_y = height_tex   
        if orientation == 'portrait':
            if width_tex < height_tex:
                # Portrait texture
                if height_tex / width_tex > 1.5:
                    # texture is 16:9
                    if self.aspect_ratio == '4:3':
                        crop_siz_y = width_tex * aspect   #1
                        crop_pos_y = (height_tex - crop_siz_y) // 2
                else:
                    # texture is 4:3
                    if self.aspect_ratio == '16:9':
                        crop_siz_x = height_tex // aspect
                        crop_pos_x = (width_tex - crop_siz_x) // 2
            else:
                # Landscape texture
                crop_siz_x = height_tex // aspect
                crop_pos_x = (width_tex - crop_siz_x) // 2
        else: # Landscape
            if width_tex < height_tex:
                # Portrait texture
                crop_siz_y = width_tex // aspect
                crop_pos_y = (width_tex - crop_siz_y) // 2
            else:
                # Landscape texture
                if width_tex / height_tex > 1.5:
                    # texture is 16:9
                    if self.aspect_ratio == '4:3':
                        crop_siz_x = height_tex * aspect
                        crop_pos_x = (width_tex - crop_siz_x) // 2
                else:
                    # texture is 4:3
                    if self.aspect_ratio == '16:9':
                        crop_siz_y = width_tex // aspect
                        crop_pos_y = (height_tex - crop_siz_y) // 2
        return [ int(crop_pos_x), int(crop_pos_y),
                 int(crop_siz_x), int(crop_siz_y)]

    @mainthread
    def canvas_text(self,text):
        label = CoreLabel(font_size = sp(16))
        label.text = text
        label.refresh()        
        if label.texture:
            pos = [self.view_pos[0] +\
                   (self.view_size[0] - label.texture.size[0]) / 2,
                   self.view_pos[1] + self.view_size[1] / 2]            
            with self.canvas:
                Color(0.6,0.6,0.6,1)
                Rectangle(size = self.view_size, pos = self.view_pos)
                Color(1,0,0,1)
                Rectangle(size=label.texture.size,
                          pos=pos,
                          texture=label.texture)
            



# An implementation of Android CameraX called from a Kivy Preview widget.
#
# About CameraX:
#    https://developer.android.com/training/camerax
# Tested devices:
#    https://developer.android.com/training/camerax/devices
#
# Source
# https://github.com/Android-for-Python/Camera4Kivy/preview_camerax.py
#

from kivy.clock import Clock, mainthread
from kivy.graphics import Fbo, Callback, Rectangle, Rotate, Scale, Translate,\
    Color
from kivy.graphics.texture import Texture
from kivy.logger import Logger

from datetime import datetime
from os.path import exists, join
from os import mkdir, remove
from pathlib import Path
from threading import Thread

from gestures4kivy import CommonGestures
from camera4kivy.preview_common import PreviewCommon

from android.storage import app_storage_path, primary_external_storage_path
from android.runnable import run_on_ui_thread
from android import mActivity, api_version
from jnius import autoclass, PythonJavaClass, java_method

GL_TEXTURE_EXTERNAL_OES = autoclass(
    'android.opengl.GLES11Ext').GL_TEXTURE_EXTERNAL_OES
Environment = autoclass('android.os.Environment')
CameraX = autoclass('org.kivy.camerax.CameraX')
if api_version >= 29:
    ContentValues = autoclass('android.content.ContentValues')
    MediaStoreMediaColumns =\
        autoclass('android.provider.MediaStore$MediaColumns')
    MediaStoreImagesMedia =\
        autoclass('android.provider.MediaStore$Images$Media')
    FileInputStream = autoclass('java.io.FileInputStream')
    FileUtils        = autoclass('android.os.FileUtils')
            
class PreviewCameraX(PreviewCommon, CommonGestures):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._camera = None
        self.enable_zoom_gesture = False
        self.enable_focus_gesture  = False
        self.block_pipeline = True
        self._fbo = None
    
    ##############################
    # Lifecycle events
    ##############################

    def connect_camera(self,
                       enable_photo = True,
                       enable_video = True,
                       enable_analyze_imageproxy  = False,
                       camera_id = 'back',
                       optimize = 'quality',
                       sensor_resolution = None,
                       default_flash = 'off',
                       default_zoom = 0.5,
                       enable_zoom_gesture = True,
                       enable_focus_gesture = True,   
                       data_format = 'yuv420',
                       filepath_callback = None,
                       analyze_proxy_callback = None,
                       analyze_callback = None,
                       canvas_callback = None,
                       **kwargs):

        self._camera = None
        self._update_ev = None
        self._name_pipe = []
        self.texture_size = []
        self.rotation = 0
        self.capture_in_progress = False

        # uniform case
        self.flash_state = default_flash.lower()
        data_format = data_format.lower()
        optimize = optimize.lower()

        self.canvas_callback = canvas_callback
        self.set_filepath_callback(filepath_callback)
        self.set_facing(camera_id)
        self.set_resolution(sensor_resolution)
        self.enable_data = enable_analyze_imageproxy
        
        # flash
        if self.flash_state not in ['on','off','auto']:
            self.flash_state = 'off'

        # optimize
        if optimize not in ['latency','quality']:
            optimize = 'quality'

        # zoom and focus
        default_zoom = min(max(default_zoom,0),1)
        self.enable_zoom_gesture = enable_zoom_gesture
        self.enable_focus_gesture = enable_focus_gesture

        # Analyse Image format
        if data_format not in ['rgba', 'yuv420']:
            data_format = 'yuv420'

        self._analyze_callback = analyze_callback
        self._analyze_proxy_callback = analyze_proxy_callback

        
        # These Java callbacks will execute in Java Main Thread
        self.cb_wrapper = CallbackWrapper(self._filename_callback,
                                          self._analyze_image_proxy,
                                          self._configure_pipeline)

       # Create an Android camera with the required behavior
        self._camera = CameraX(
            enable_photo,
            enable_video,
            self.enable_data,
            self.facing,
            self._sensor_resolution,
            self.aspect_ratio,
            self.cb_wrapper,
            self.flash_state,
            optimize,
            default_zoom,
            data_format)

        # check camerax_provider version
        # Set in org/kivy/camerax/CameraX.java
        latest = '0.0.3'  
        try:
            if self._camera.providerVersion() < latest:
                Logger.warning('Update camerax_provider to the latest version, this is ' + latest)
        except:
            Logger.warning('Update camerax_provider to the latest version, this is ' + latest)

        # Configure the camera for the Kivy view port
        self._configure_camera(True)

    def disconnect_camera(self):
        self.destroy_camera()
        self.block_and_clear_pipeline() 
        
    @run_on_ui_thread
    def destroy_camera(self):
        self.stop_capture_video()
        self._deschedule_pipeline()
        if self._camera:
            self._camera.unbind_camera()
            self._camera = None
                
    # configure camera
    def _configure_camera(self, start):
        self.configure_viewport()
        if self._camera:
            self._camera.setViewPort(self.view_size)
            self._camera.startCamera()
        else:
            self.canvas.clear()
            with self.canvas:
                Color(1,1,1,1)
                Rectangle(size = self.view_size, pos = self.view_pos)

    # Device Rotate
    def on_size(self, instance, size):
        if self._camera:
            self.stop_capture_video()
            self._configure_camera(False)
            
    ##################################
    # Parse options
    ##################################

    def set_facing(self, facing):
        facing = facing.lower()
        if facing == '0':
            facing = 'back'
        elif facing == '1':
            facing = 'front'
        elif facing not in ['back','front']:
            facing = 'back'
        self.facing = facing

    ##############################
    # Preview Widget Touch Events
    ##############################

    # CommonGestures Touch Events
    # tap for focus
    def cgb_primary(self, touch, x, y):
        if self._camera and self.enable_focus_gesture:
            self.focus(x, y)

    # pinch/spread for zoom
    def cgb_zoom(self, touch0, touch1, x, y, scale):
        if self._camera and self.enable_zoom_gesture:
            self.zoom_delta(scale)

    ##############################
    # User events
    ##############################

    def capture_photo(self, location = '',  subdir = '', name = ''):
        if self._camera:
            self.capture_in_progress = True
            self._set_location(location)
            subdir = self._default_subdir_android(subdir)
            name = self._default_file_name(name, '.jpg')
            if self.file_storage:
                self._name_pipe.append(join(subdir, name))
            self._camera.capture_photo(subdir, name, self.file_storage)

    def capture_video(self, location = '', subdir = '', name = ''):
        if self._camera:
            self.capture_in_progress = True
            self._set_location(location)
            subdir = self._default_subdir_android(subdir)
            name = self._default_file_name(name,'.mp4')
            if self.file_storage:
                self._name_pipe.append(join(subdir, name))
            self._camera.capture_video(subdir, name, self.file_storage)

    def stop_capture_video(self):
        if self._camera:
            self._camera.stop_capture_video()

    def capture_screenshot(self, location = '.', subdir = '', name = ''):
        view_crop = self.screenshot_crop()
        self._set_location(location)
        subdir = self._default_subdir_android(subdir)
        name = self._default_file_name(name, '.jpg')
        tex = self.export_as_image().texture.get_region(*view_crop)
        path = join(subdir, name)
        if self.file_storage:
            # local or, shared and api<=29
            tex.save(path, flipped = True)
            if self.callback:
                self.callback(path)
        else:
            # MediaStore
            cache = self.cache_path()
            if cache:
                # write to cache
                cachefile = join(cache, name)
                tex.save(cachefile, flipped = True)
                # create MediaStore entry
                cv = ContentValues()
                cv.put(MediaStoreMediaColumns.DISPLAY_NAME, name)
                cv.put(MediaStoreMediaColumns.MIME_TYPE, 'image/jpeg')  
                cv.put(MediaStoreMediaColumns.RELATIVE_PATH, subdir)
                root_uri = MediaStoreImagesMedia.getContentUri('external')
                context = mActivity.getApplicationContext()    
                uri = context.getContentResolver().insert(root_uri, cv)
                # copy cache file to MediaStore
                rs = FileInputStream(cachefile)
                ws = context.getContentResolver().openOutputStream(uri)
                FileUtils.copy(rs,ws)    
                ws.flush()
                ws.close()
                rs.close()                
                remove(cachefile)
                if self.callback:
                    self.callback(path)
    

    # Select back, front camera
    def select_camera(self, facing):
        if self._camera:
            facing = facing.lower()
            if facing == 'toggle':
                if self.facing == 'back':
                    self.facing = 'front'
                else:
                    self.facing = 'back'
            elif facing == 'front' or facing == '1':
                self.facing = 'front'
            else:
                self.facing = 'back'

            # may have to wait for a capture to complete
            if not self.capture_in_progress:
                self.block_and_clear_pipeline()
                self.do_select_camera()
            else:
                self.stop_capture_video()
                self._facing_ev = Clock.schedule_interval(
                    self.can_select_camera, 1 / 30)
            facing = self.facing
        return facing

    def can_select_camera(self,dt):
        if not self.capture_in_progress:
            self.block_and_clear_pipeline()
            self.do_select_camera()
            Clock.unschedule(self._facing_ev)

    @run_on_ui_thread
    def do_select_camera(self):
        self._camera.select_camera(self.facing)
            
    # Sequence flash : off, on, auto, ...
    def flash(self, state = None):
        # None, auto sequence, 0 ->2 set state
        if self._camera:
            if state == None:
                if self.flash_state == 'off':
                    self.flash_state = 'on'
                elif self.flash_state == 'on':
                    self.flash_state = 'auto'
                else:
                    self.flash_state = 'off'
            elif state in ['off', 'on', 'auto']:
                self.flash_state = state
            self.flash_state = self._camera.flash(self.flash_state)
            return self.flash_state
        return "off"

    def torch(self, state = None):
        if self._camera:
            if state in ['off', 'on']:
                try:
                    return self._camera.torch(state)
                except:
                    Logger.warning('Update camerax_provider to >= 0.0.3')
                    return 'off'
        else:
            return 'off'
                
    # if enable_focus_gesture == True, then this is called by a tap gesture
    def focus(self, x, y):
        if self._camera:
            self._camera.focus(x, y)

    # if enable_zoom_gesture == True, then this called by pinch/spread gesture
    def zoom_delta(self, delta_scale):
        if self._camera:
            self._camera.zoom(delta_scale, False)

    def zoom_abs(self, scale):
        if self._camera:
            self._camera.zoom(scale, True)

    ##############################
    # Create Preview Pipeline
    ##############################
        
    def _create_texture(self, size):
        self._camera_texture = Texture(width = size[0],
                                       height= size[1],
                                       target=GL_TEXTURE_EXTERNAL_OES,
                                       colorfmt='rgba')
        return int(self._camera_texture.id)

    def _create_fbo(self, texture_size, rotation):
        long_edge = max(texture_size)
        short_edge = min(texture_size)
        origin    = (texture_size[0]/2, texture_size[1]/2)
        translate = 0
        scalex = 1
        scaley = 1

        if rotation == 90:
            translate = -(long_edge - short_edge) /2
        elif rotation == 270:
            translate = (long_edge - short_edge) /2

        if texture_size[0] < texture_size[1]:
            translate = -translate
            scalex = -scalex
            scaley = -scaley

        if rotation in [90 , 270]:            
            fbo_size = (texture_size[1],texture_size[0])
        else:
            fbo_size = texture_size

        self._fbo = Fbo(size=fbo_size)
        self._fbo.shader.fs = '''
            #extension GL_OES_EGL_image_external : require
            $HEADER$
            uniform samplerExternalOES texture1;
            void main()
            {
                gl_FragColor = texture2D(texture1, tex_coord0);
            }
        '''
        
        with self._fbo.before:
            Rotate(origin = origin, angle = 360 - rotation, axis = (0, 0, 1))
            Translate(translate, translate)
            Scale(scalex, scaley, 1, origin = origin )
            Rectangle(size = texture_size)

        with self._fbo:
            self._camera_texture_cb = Callback(lambda instr:
                                               self._camera_texture.bind)

    # Run on Kivy main thread because required by FBO.
    @mainthread
    def _create_pipeline(self, texture_size, rotation):
        id = self._create_texture(texture_size)
        self._create_fbo(texture_size, rotation)
        self._camera.setTexture(id,texture_size)
        self._schedule_pipeline()

    ##############################
    # Fill Preview Pipeline
    ##############################
            
    def block_and_clear_pipeline(self):
        self.block_pipeline = True
        if self._camera and self._fbo and self._fbo.texture:
            tex_size = self._fbo.texture.size
            buf = bytes([255] * tex_size[0] * tex_size[1] * 4)
            self._fbo.texture.blit_buffer(buf, colorfmt='rgba',
                                          bufferfmt='ubyte')
            self._analyze_texture()
            self._update_canvas() 

    def _schedule_pipeline(self):
        self._deschedule_pipeline()
        if self._camera and self._camera_texture and self._fbo and\
           self._fbo.texture:
            self._set_surface_provider(True)
            self.block_pipeline = False
            self._update_ev = Clock.schedule_interval(self._update_pipeline,
                                                      1 / 30)

    def _deschedule_pipeline(self):
        if self._update_ev is not None:
            self._set_surface_provider(False)
            self._update_ev.cancel()
            self._update_ev = None

    def _update_pipeline(self, dt):
        if self._camera.imageReady() and self._fbo and not self.block_pipeline:
            self._camera_texture_cb.ask_update() 
            self._fbo.draw()  
            self._analyze_texture()
            self._update_canvas() 

    # Run on UI thread because required by CameraX
    @run_on_ui_thread
    def _set_surface_provider(self, enable):
        self._camera.setSurfaceProvider(enable)

    # Run on mainthread because required by Kivy canvas
    @mainthread
    def _update_canvas(self):
        if self._fbo:
            tex = self._fbo.texture.get_region(*self.crop)

            # moved from create_fbo
            if self.facing == 'front':
                view_size = (-self.view_size[0], self.view_size[1])
                view_pos = (self.view_pos[0] + self.view_size[0],
                            self.view_pos[1])
            else:
                view_size = self.view_size
                view_pos  = self.view_pos
        
            self.canvas.clear()
            with self.canvas:
                Color(1,1,1,1)
                Rectangle(texture= tex, size = view_size, pos = view_pos)
                if self.canvas_callback:
                    self.canvas_callback(tex, view_size, view_pos)

    #######################################
    # Storage Location
    #######################################

    def _set_location(self, location):
        storage = location.lower()
        if storage not in ['private', 'shared']:
            storage = 'shared'
        self.private_storage = storage == 'private'
        self.file_storage = self.private_storage or api_version < 29

    def _default_location(self):
        if self.private_storage:
            root = join(app_storage_path(),Environment.DIRECTORY_DCIM)
            if not exists(root):
                mkdir(root)
        else:
            if api_version < 29:
                root = join(primary_external_storage_path(),
                            Environment.DIRECTORY_DCIM,
                            self._app_name())
                if not exists(root):
                    mkdir(root)
            else:
                root = join(Environment.DIRECTORY_DCIM, self._app_name())
        return root
                
    def _default_subdir_android(self, subdir=''):
        root = self._default_location()
        if not subdir:
            # Today's date
            subdir = datetime.now().strftime("%Y_%m_%d")
        path = join(root,subdir)
        if self.private_storage or api_version < 29:
            if not exists(path):
                mkdir(path)
        return path

    def _app_name(self):
        context = mActivity.getApplicationContext()
        appinfo = context.getApplicationInfo()
        if appinfo.labelRes:
            name = context.getString(appinfo.labelRes)
        else:
            name = appinfo.nonLocalizedLabel.toString()
        return name

    def cache_path(self):
        context = mActivity.getApplicationContext()
        cache =  context.getExternalCacheDir()
        return str(cache.toString())

    #######################################
    # Callbacks
    #######################################

    # Runs in Java Main Thread
    def _configure_pipeline(self, croprect, resolution, rotation):
        if rotation in [ 90, 270]:  
            self.crop = [croprect.top,   croprect.left,
                         croprect.bottom - croprect.top,
                         croprect.right - croprect.left]
        else:  
            self.crop = [croprect.left,   croprect.top,
                         croprect.right - croprect.left,
                         croprect.bottom - croprect.top]
        texture_size = [resolution.getWidth(), resolution.getHeight()]
        self.texture_size = texture_size
        self.tscale = self.view_size[1] / self.crop[3]
        self.rotation = rotation
        self._create_pipeline(texture_size, rotation)

    # Runs in some Java thread
    def _filename_callback(self, file_id):
        if not file_id:
            # The callback returns "" for non-MediaStore saves
            if self._name_pipe:
                file_id = self._name_pipe[0]
                self._name_pipe = self._name_pipe[1:]
        self.capture_in_progress = False
        if self.callback:
            self.callback(str(file_id))

    def _analyze_texture(self):
        if not self.enable_data and self._fbo and self._analyze_callback:
            tex = self._fbo.texture.get_region(*self.crop)
            self._analyze_callback(tex, self.view_pos,
                                   self.tscale, self.facing=='front')

    def _analyze_image_proxy(self, image_proxy):
        if self.enable_data and self._analyze_proxy_callback:
            if self.rotation in [0, 180]:
                tscale = self.view_size[1] / image_proxy.getHeight()
            else:
                tscale = self.view_size[1] / image_proxy.getWidth()
            self._analyze_proxy_callback(image_proxy, self.view_pos,
                                         tscale, self.facing == 'front',
                                         self.rotation)

       
class CallbackWrapper(PythonJavaClass):
    __javacontext__ = 'app'
    __javainterfaces__ = ['org/kivy/camerax/CallbackWrapper']

    def __init__(self, callback0, callback1, callback2):
        super().__init__()
        self.callback0 = callback0
        self.callback1 = callback1
        self.callback2 = callback2

    @java_method('(Ljava/lang/String;)V')        
    def callback_string(self, filepath):
        if self.callback0:
            self.callback0(filepath)

    @java_method('(Landroidx/camera/core/ImageProxy;)V')        
    def callback_image(self, image):
        if self.callback1:
            self.callback1(image)

    @java_method('(Landroid/graphics/Rect;Landroid/util/Size;I)V')  
    def callback_config(self, croprect, resolution, rotation):
        if self.callback2:
            self.callback2(croprect, resolution, rotation)
            



'''
PiCamera2 Camera: Implement CameraBase with PiCamera2
'''

__all__ = ('CameraPiCamera2', )
import io

from kivy.logger import Logger
from kivy.clock import Clock, mainthread
from kivy.graphics.texture import Texture
from kivy.graphics import Color, Rectangle, Rotate, Translate, Fbo, BindTexture

from picamera2 import Picamera2
from picamera2.previews import NullPreview
from picamera2.encoders import H264Encoder, Quality 
from picamera2.outputs import Output
from picamera2.request import _MappedBuffer

import numpy as np
from os import environ
from PIL import Image
from . import CameraBase

import signal
import subprocess
import prctl

# Libcamera defaults to INFO, too verbose
environ['LIBCAMERA_LOG_LEVELS'] = 'ERROR'

#################################
# Picamera2 Sensor Interface
#################################

class SensorInterface(NullPreview):

    def __init__(self):
        super().__init__()
        self.mute = False
        self.y = None
        self.u = None
        self.v = None
        self.mjpeg = None
        self.stream_size = ()

    # Sync event loops
    ###################
    @mainthread
    def sync_yuv(self,y,u,v):
        self.y = y
        self.u = u
        self.v = v

    @mainthread
    def sync_mjpeg(self,mjpeg):
        self.mjpeg = mjpeg
    
    # Request Handlers
    ###################
    def handle_request(self, picam2):
        if self.picam2.display_stream_name:
            camera_config = self.picam2.camera_config
            self.display_stream_name = self.picam2.display_stream_name
            stream_config = camera_config[self.display_stream_name]
            self.stream_fmt = stream_config["format"]
            self.stream_size = stream_config['size']
        picam2.process_requests(self)

    def render_request(self, request):
        try:
            # The added latency due to array manipulation in some of
            # these, occurs in the Picamera2 thread and is much less
            # than sample period of the Kivy thread. Nyquist is happy.
            with _MappedBuffer(request,self.display_stream_name) as mm:
                if self.stream_fmt == 'YUV420':
                    size = len(mm)
                    end_y = size * 2 // 3
                    end_u = end_y + end_y // 4
                    y = bytes(mm[:end_y])
                    u = bytes(mm[end_y:end_u])
                    v = bytes(mm[end_u:])
                    self.sync_yuv(y,u,v)
                                
                elif self.stream_fmt == 'MJPEG':
                    self.sync_mjpeg(bytes(mm))
                    
                elif self.stream_fmt and not self.mute:
                    self.mute = True
                    Logger.error(
                        "Picamera2 SensorInterface unsupported format " +\
                        self.stream_fmt)

        except Exception as e:
            Logger.error("Picamera2 SensorInterface\n" + str(e))


####################################
# FfmpegOutput with rotate metadata
####################################

class FfmpegOutputPlus(Output):

    def __init__(self, output_filename, audio=False, audio_device="default",
                 audio_sync=-0.3, audio_samplerate=48000, audio_codec="aac",
                 audio_bitrate=128000, pts=None, rotate = None):
        super().__init__(pts=pts)
        self.output_filename = output_filename
        self.audio = audio
        self.audio_device = audio_device
        self.audio_sync = audio_sync
        self.audio_samplerate = audio_samplerate
        self.audio_codec = audio_codec
        self.audio_bitrate = audio_bitrate
        self.rotate = rotate

    def start(self):
        general_options = ['-loglevel', 'warning', '-y']  
        video_input = ['-use_wallclock_as_timestamps', '1',
                       '-thread_queue_size', '32',  
                       '-i', '-']
        if self.rotate:  #-map_metadata 0 -metadata:s:v rotate="90"
            video_input = video_input + ['-map_metadata', '0', '-metadata:s:v',
                                         'rotate='+str(self.rotate)]
        video_codec = ['-c:v', 'copy']
        audio_input = []
        audio_codec = []
        if self.audio:
            audio_input = ['-itsoffset', str(self.audio_sync),
                           '-f', 'pulse',
                           '-sample_rate', str(self.audio_bitrate),
                           '-thread_queue_size', '512',  
                           '-i', self.audio_device]
            audio_codec = ['-b:a', str(self.audio_bitrate),
                           '-c:a', self.audio_codec]

        command = ['ffmpeg'] + general_options + audio_input + video_input + \
            audio_codec + video_codec + self.output_filename.split()

        self.ffmpeg = subprocess.Popen(command, stdin=subprocess.PIPE,
                                       preexec_fn=lambda: prctl.set_pdeathsig(signal.SIGKILL))
        super().start()

    def stop(self):
        super().stop()
        if self.ffmpeg is not None:
            self.ffmpeg.stdin.close()  # FFmpeg needs this to shut down tidily
            self.ffmpeg.terminate()
            self.ffmpeg = None

    def outputframe(self, frame, keyframe=True, timestamp=None):
        if self.recording:
            self.ffmpeg.stdin.write(frame)
            self.ffmpeg.stdin.flush()  # forces every frame to get timestamped individually
            self.outputtimestamp(timestamp)

            
            

#################################
# Picamera2 Camera Lifecycle
#################################

class CameraPi2():

    def __init__(self):
        super().__init__()
        self.sensor = None
        self.picam2 = None
        self.base_scaler_crop = None
        self.scaler_crop = None
        self.zoom_level = None
        self._rotate = 0
        self.video_recording = False
        self.audio = False
        self.is_usb = False

    # Start
    # Choose sensor, configure pc2
    ###############################

    def start(self, index):
        self.zoom_level = 1
        self.previous_fmt = ''
        self.previous_size = (0,0)
        self.previous_tsize = (0,0)
        self.fbo = None
        
        # Get info about this camera
        num_cameras = len(Picamera2.global_camera_info())
        if num_cameras == 0:
            Logger.error('C4k Picamera2: No camera found.')
            return
        if index <0 or index >= num_cameras:
            Logger.warning('C4k Picamera2: Requested camera '+ str(index) +\
                           ' not found, using ' + str(num_cameras -1) + ' .')
            index = num_cameras -1

        # initialize 
        Id = Picamera2.global_camera_info()[index]['Id']
        self.picam2 = Picamera2(index)
        if 'i2c' in Id.lower():
            self.is_usb = False
            self.create_picam_configurations(index)
        else:
            self.is_usb = True
            self.create_usb_configurations()
        self.base_scaler_crop = self.crop_limits 
        self.scaler_crop = self.crop_limits
        self.picam2.configure(self.preview_config)
        self.sensor= SensorInterface()
        self.picam2.start_preview(self.sensor)
        self.picam2.start()

    def create_usb_configurations(self):
        self.crop_limits = None
        self.preview_config = self.picam2.create_preview_configuration(
            {"format": "MJPEG"})
        self.photo_config = self.picam2.create_still_configuration(
            {"format": "MJPEG"})
        self.video_config = self.picam2.create_video_configuration(
            {"format": "MJPEG"})  # Not supported

    def create_picam_configurations(self, index):
        # Sensor configuration
        size_s = (0,0)
        wide = self._context.aspect_ratio == '16:9'
        for m in self.picam2.sensor_modes:
            # Raspberry camera resolution is also field of view.
            # get highest sensor resoluton for this framerate
            # so framerate will set field of view depending on camera
            #
            # Because video players don't support crop metadata well,
            # we can't crop from 4:3 to 16:9
            # So use native 16:9 and shift in fbo
            if 'fps' in m and 'size' in m and 'bit_depth' in m:
                fps = m['fps']
                size = m['size']
                bits = m['bit_depth']
                if fps >= self._framerate and bits == 8:   
                    if not wide and size[0]/size[1] < 1.5:
                        if size[0] > size_s[0]:
                            size_s = size
                            self.crop_limits = m['crop_limits']
                    elif wide and size[0]/size[1] >= 1.5:
                        if size[0] > size_s[0]:
                            size_s = size
                            self.crop_limits = m['crop_limits']

        if not size_s[0]:
            Logger.error('No sensor found in supporting ' +\
                         self.aspect_ratio + ' and ' +\
                         str(self._framerate) + ' fps.')
            return

        # Stream sizes for each configuration
        def align(edge, val):
            return val * round(edge / val)

        dw = align(self._resolution[0] , 64)
        if wide:
            dh = align(self._resolution[0] * 9 / 16, 64)
            vh = 720
        else:
            dh = align(self._resolution[1], 64)
            vh = 960
        main = {"size": (align(size_s[0], 16), align(size_s[1], 16)) }
        preview_lores = {"size": (dw, dh)}
        video_lores = {"size": (1280, vh)}
        
        # Configurations
        self.preview_config = self.picam2.create_preview_configuration(
            main = main,
            lores = preview_lores,
            display = 'lores')
        self.photo_config = self.picam2.create_still_configuration(
            main = main)
        self.video_config = self.picam2.create_video_configuration(
            main = main,
            lores = video_lores,
            encode = 'lores',
            display = 'lores')


    # Stop
    ###############################
    
    def stop(self):
        if self.sensor:
            self.sensor.stop()
        if self.picam2:
            self.picam2.close()
        self.sensor = None
        self.picam2 = None
        self.photo_config = None
        self.video_config = None

    # Display Update
    ###############################

    def update(self):
        ss = self.sensor
        if ss and ss.y:
            return self._yuv_to_rgba('YUV420', ss.y, ss.u, ss.v,
                                     ss.stream_size, self._resolution)
        elif ss and ss.mjpeg:
            img = Image.open(io.BytesIO(ss.mjpeg))
            img = img.convert('RGBA')
            img = img.rotate(self._rotate)
            img = img.resize(self._resolution)
            return img.tobytes()
        return None            

    # Zoom and Drag events
    ###############################

    def zoom(self, scale):
        if self.picam2 and self.base_scaler_crop:
            self.zoom_level /= scale   # wheel on pi is backwards
            self.set_zoom()

    def set_zoom(self):
        max_zoom = 7.0
        if self.zoom_level < 1:
            self.zoom_level = 1.0
        if self.zoom_level > max_zoom:
            self.zoom_level = max_zoom
        factor = 1.0 / self.zoom_level
        full_img = self.base_scaler_crop
        center = (self.scaler_crop[0] + self.scaler_crop[2] // 2,
                  self.scaler_crop[1] + self.scaler_crop[3] // 2)
        w = int(factor * full_img[2])
        h = int(factor * full_img[3])
        x = full_img[0] + center[0] - w // 2
        y = full_img[1] + center[1] - h // 2
        self.limit_and_save([x, y, w, h])

    def drag(self, dx, dy):
        if self.picam2 and self.base_scaler_crop:
            full_img = self.base_scaler_crop
            w = self.scaler_crop[2]
            h = self.scaler_crop[3]
            x = self.scaler_crop[0] + int(full_img[2] * dx)
            y = self.scaler_crop[1] + int(full_img[3] * dy)
            self.limit_and_save([x, y, w, h])

    def limit_and_save(self,new_scaler_crop):
        full_img = self.base_scaler_crop
        new_scaler_crop[1] = min(max(new_scaler_crop[1], full_img[1]),
                                 full_img[1] + full_img[3] - new_scaler_crop[3])
        new_scaler_crop[0] = min(max(new_scaler_crop[0], full_img[0]),
                                 full_img[0] + full_img[2] - new_scaler_crop[2])
        self.scaler_crop = tuple(new_scaler_crop)
        self.picam2.controls.ScalerCrop = self.scaler_crop


    # Photo start/stop capture
    ###############################

    def capture_file(self, file_output, callback):
        request = self.picam2.capture_request()
        size = request.config['main']['size']
        with _MappedBuffer(request,'main') as pixels:
            if self.is_usb:
                img = Image.open(io.BytesIO(pixels))
            else:
                img = Image.frombytes('RGB', size, bytes(pixels))
        request.release()
        if self._rotate in [90,270]:
            size = size[::-1]
        crop = self._context.crop_for_aspect_orientation(size[0],
                                                         size[1])
        bottom = crop[3] + crop[1]
        right = crop[2] + crop[0]
        img = img.rotate(self._rotate, expand = True)
        img = img.crop((crop[0], crop[1], right, bottom))
        with open(file_output, 'wb') as fp:
            img.save(fp)
        if callback:
            callback(file_output)
            
    # picam2.switch_mode loses ScalarCrop
    def switch_config(self, new_config):
        self.picam2.stop()
        self.picam2.configure(new_config)
        if self.is_usb:
            self.picam2.start()
        else:
            self.picam2.controls.ScalerCrop = self.scaler_crop
            self.picam2.start()
            self.picam2.controls.ScalerCrop = self.scaler_crop
        
    def photo(self, path, callback):
        if self.picam2 and self.sensor and not self.video_recording:
            self.switch_config(self.photo_config)
            self.capture_file(path, callback)
            self.switch_config(self.preview_config)

    # Video start/stop
    ###############################

    def video_start(self, filepath, callback):
        if self.is_usb:
            Logger.error('Camera4Kivy, USB video recording not supported.')
            return
        self.video_filepath = filepath
        self.video_callback = callback
        if self.picam2 and self.sensor:
            self.video_recording = True
            self.picam2.switch_mode(self.video_config)
            encoder = H264Encoder()
            output = FfmpegOutputPlus(filepath, rotate = self._rotate,
                                      audio= self.audio, audio_sync = 0)
            self.picam2.start_encoder(encoder, output)

    def video_stop(self):
        if self.is_usb:
            return
        self.picam2.stop_encoder()
        self.picam2.switch_mode(self.preview_config)
        self.video_recording = False
        if self.video_callback:
            self.video_callback(self.video_filepath)

    # YUV reformatting
    ###############################

    YUV_RGB_FS = '''
    $HEADER$
    uniform sampler2D tex_y;
    uniform sampler2D tex_u;
    uniform sampler2D tex_v;
    mat3 YUV2RGB_JPEG = mat3(1.0,   1.0,   1.0  ,
                             0.0, -0.344, 1.772,
                             1.402, -0.714, 0.0);
    mat3 YUV2RGB_SMPTE170M = mat3(1.164, 1.164, 1.164,
                                  0.0, -0.392, 2.017,
                                  1.596, -0.813, 0.0);
    mat3 YUV2RGB_REC709 = mat3(1.164, 1.164, 1.164,
                               0.0, -0.213, 2.112,
                               1.793, -0.533, 0.0);
    void main(void) {
        vec3 yuv;
        yuv.r = texture2D(tex_y, tex_coord0).r;
        yuv.g = texture2D(tex_u, tex_coord0).r -0.5;
        yuv.b = texture2D(tex_v, tex_coord0).r -0.5;
        gl_FragColor = vec4(YUV2RGB_JPEG * yuv, 1.0);
    }
    '''

    def _yuv_to_rgba(self, fmt, y, u, v, size, tsize):
        if self._context.aspect_ratio == '16:9':
            isize = [tsize[0], round(tsize[0] * 9 / 16)]
            translate = (tsize[1] - isize[1]) // 2
        else:
            isize = tsize
            translate = 0
        origin = (tsize[0]//2, tsize[1]//2)
        if fmt == 'YUV420':
            uv_size = (size[0]//2, size[1]//2 )
        else:
            uv_size = (size[0]//2, size[1]) 
            
        if self.previous_size[0] != size[0] or self.previous_size[1] != size[1]:
            self.tex_y = Texture.create(size= size, colorfmt='luminance')
            self.tex_u = Texture.create(size= uv_size, colorfmt='luminance')
            self.tex_v = Texture.create(size= uv_size, colorfmt='luminance')

        if self.previous_tsize[0] != tsize[0] or\
           self.previous_tsize[1] != tsize[1] or\
               self.fbo == None:
            self.previous_tsize = tsize
            self.fbo = Fbo(size=tsize)   # size for bilt to self._texture
            self.fbo.texture.flip_vertical()
            with self.fbo:
                self.b_u = BindTexture(texture=self.tex_u, index=1)
                self.b_v = BindTexture(texture=self.tex_v, index=2)
                Rotate(origin = origin, angle = 360-self._rotate,
                       axis = (0, 0, 1))
                Translate(0, translate)
                self.r_y = Rectangle(size=isize, texture=self.tex_y)
            self.fbo.shader.fs = self.YUV_RGB_FS
            self.fbo['tex_y'] = 0
            self.fbo['tex_u'] = 1
            self.fbo['tex_v'] = 2
            
        if self.previous_size[0] != size[0] or self.previous_size[1] != size[1]:
            self.previous_size = size
            self.r_y.size = isize
            self.r_y.texture = self.tex_y
            self.b_u.texture = self.tex_u
            self.b_v.texture = self.tex_v
            # Repeat previous pixels to prevent flicker on change
            return bytes(self.fbo.texture.pixels)
        
        self.tex_y.blit_buffer(y, colorfmt='luminance')
        self.tex_u.blit_buffer(u, colorfmt='luminance')
        self.tex_v.blit_buffer(v, colorfmt='luminance')
        self.fbo.ask_update()
        self.fbo.draw()
        return self.fbo.texture.pixels


#################################
# Kivy Camera Provider
#################################

class CameraPiCamera2(CameraBase):
    '''Implementation of CameraBase using PiCamera2  
    '''

    def __init__(self, **kwargs):
        self._update_ev = None
        self._camera = None
        self._framerate = kwargs.get('framerate', 30)
        self.started = False
        self.fbo = None
        self._rotate = kwargs.get('rotation', 0) 
        self.audio = kwargs.get('audio', False) 
        super().__init__(**kwargs)

    # Lifecycle
    ################################

    def init_camera(self):
        self._format = 'rgba'   
        if self._camera is not None:
            self._camera.close()
        self._texture = None
        self.stopped = True
        self.fps = 1. / self._framerate

    def update(self, dt):
        if self.stopped:
            return
        if self._texture is None:
            self._texture = Texture.create(self._resolution)
            self._texture.flip_vertical()
            self._context.on_load()
        try:
            self._buffer = self._camera.update()
            if self._buffer:
                self._copy_to_gpu()
        except Exception as e:
            Logger.error('CameraPiCamera2\n' + str(e))

    def start(self): 
        if not self.started:
            self.started = True
            super().start()
            self._camera = CameraPi2()
            self._camera._resolution = self._resolution
            self._camera._framerate = self._framerate
            self._camera._context = self._context
            self._camera._rotate = self._rotate
            self._camera.audio = self.audio
            self._texture = None
            if self._update_ev is not None:
                self._update_ev.cancel()
            self._camera.start(self._index)
            self._update_ev = Clock.schedule_interval(self.update, self.fps)

    def stop(self):
        super().stop()
        self.started = False
        #self._camera = None

        if self._update_ev is not None:
            self._update_ev.cancel()
            self._update_ev = None
        if self._camera:
            self._camera.stop()
        self._texture = None
        self.fbo = None

    def photo(self,filepath, callback):
        if self._camera:
            self._camera.photo(filepath, callback)

    def video_start(self,filepath, callback):
        if self._camera:
            self._camera.video_start(filepath, callback)

    def video_stop(self):
        if self._camera:
            self._camera.video_stop()

    def zoom(self, scale):
        if self._camera:
            self._camera.zoom(scale)

    def drag(self, dx, dy):
        if self._camera:
            self._camera.drag(dx, dy)




'''
Camera
======

Core class for acquiring the camera and converting its input into a
:class:`~kivy.graphics.texture.Texture`.

.. versionchanged:: 1.10.0
    The pygst and videocapture providers have been removed.

.. versionchanged:: 1.8.0
    There is now 2 distinct Gstreamer implementation: one using Gi/Gst
    working for both Python 2+3 with Gstreamer 1.0, and one using PyGST
    working only for Python 2 + Gstreamer 0.10.
'''

__all__ = ('CameraBase', 'Camera')


from kivy.utils import platform
from kivy.event import EventDispatcher
from kivy.logger import Logger
from .. import core_select_lib


class CameraBase(EventDispatcher):

    def __init__(self, **kwargs):
        kwargs.setdefault('stopped', False)
        kwargs.setdefault('index', 0)
        kwargs.setdefault('context', None)
        self.stopped = kwargs.get('stopped')
        self._resolution = kwargs.get('resolution')
        self._index = kwargs.get('index')
        self._context = kwargs.get('context')
        self._buffer = None
        self._format = 'rgb'
        self._texture = None
        self.capture_device = None
        super().__init__()
        self.init_camera()
        #if not self.stopped and not self._context:
        #    self.start()

    def _get_texture(self):
        return self._texture
    
    texture = property(lambda self: self._get_texture(),
                       doc='Return the camera texture with the latest capture')

    def init_camera(self):
        '''Initialise the camera (internal)'''
        pass

    def start(self):
        '''Start the camera acquire'''
        self.stopped = False

    def stop(self):
        '''Release the camera'''
        self.stopped = True

    def _update(self, dt):
        '''Update the camera (internal)'''
        pass

    def _copy_to_gpu(self):
        '''Copy the the buffer into the texture'''
        if self._texture is None:
            Logger.debug('Camera: copy_to_gpu() failed, _texture is None !')
            return
        self._texture.blit_buffer(self._buffer, colorfmt=self._format) 
        self._buffer = None
        if self._context:
            self._context.on_texture()
        else:
            self.dispatch('on_texture')

    #def on_texture(self):
    #    pass

    #def on_load(self):
    #    pass

import kivy
import importlib

def select_provider(category, llist, base='kivy.core'):
    category = category.lower()
    basemodule = category
    for option, modulename, classname in llist:
        if option not in kivy.kivy_options[category]:
            continue
        try:
            name = '{0}.{1}.{2}'.format(base, basemodule, modulename)
            mod = importlib.__import__(name=name,
                                       globals=globals(),
                                       locals=locals(),
                                       fromlist=[modulename], level=0)
            cls = mod.__getattribute__(classname)

            Logger.info('{0}: Provider: {1}'.format(category.capitalize(),
                                                    option))
            return cls

        except Exception as e:
            pass
        
    if platform not in ['android']:
        Logger.warning('{0}: No Provider found.'.format(category.capitalize()))
    
providers = ()

if platform in ['macosx', 'ios']:
    providers += (('avfoundation', 'camera_avfoundation',
                   'CameraAVFoundation'), )
    Camera = core_select_lib('camera', (providers))
    providers = ()
elif platform == 'win':
    providers += (('opencv', 'camera_opencv', 'CameraOpenCV'), )
    providers += (('gi', 'camera_gi', 'CameraGi'), )
elif platform == 'android':
    pass
else:
    providers += (('picamera', 'camera_picamera2', 'CameraPiCamera2'), )
    providers += (('picamera', 'camera_picamera', 'CameraPiCamera'), )
    providers += (('gi', 'camera_gi', 'CameraGi'), )
    providers += (('opencv', 'camera_opencv', 'CameraOpenCV'), )

if providers:
    Camera = select_provider('camera', (providers),
                             base='camera4kivy.based_on_kivy_core')






    



from kivy.core.window import Window
from kivy.uix.widget import Widget
from kivy.utils import platform

from os import mkdir
from os.path import exists, join
from pathlib import Path
from datetime import datetime
from inspect import ismethod, signature

class PreviewCommon(Widget):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._camera = None
        self._camera_texture = None
        self.view_size = (10, 10)
        self.view_pos = (0, 0)
        self.tex_crop = (0, 0, 10, 10)
        self.tscale = 1
        self.orientation = 'same'
        self.aspect_ratio = '4:3'
        self.callback = None
        self._sensor_resolution = []
        
    #############################################
    # Parse Arguments
    #############################################

    def set_aspect_ratio(self, aspect_ratio):
        if aspect_ratio in ['4:3', '16:9']:
            self.aspect_ratio = aspect_ratio

    def set_orientation(self, orientation):
        orientation = orientation.lower()
        if orientation in ['landscape', 'portrait','same','opposite']:
            self.orientation = orientation

    def set_rotation(self, rotation):
        self._sensor_rotation = 0
        if rotation in [0,90,180,270]:
            self._sensor_rotation = rotation

    def set_resolution(self, resolution):
        if resolution and\
           (type(resolution) is tuple or type(resolution) is list) and\
           len(resolution) == 2:
            self._sensor_resolution = (max(resolution), min(resolution))

    def set_filepath_callback(self,callback):
        if callback:
            if not ismethod(callback) or\
               len(signature(callback).parameters) !=1:
                callback = None
            self.callback = callback

    #############################################
    # Viewport
    #############################################

    def configure_viewport(self):
        orientation = self.decode_orientation()
        width_self, height_self = self.size

        if self.aspect_ratio == "4:3":
            aspect = 4/3
        else:
            aspect = 16/9
            
        if orientation == 'portrait':
            width_view  = height_self / aspect
            height_view = height_self
            if self.width < width_view:
                width_view = self.width
                height_view = self.width * aspect
            pos_x = round((self.width - width_view)/2)
            pos_y = round((self.height - height_view)/2)
            width_view = round(width_view)
            height_view = round(height_view)
        else:
            width_view = width_self
            height_view = width_self / aspect
            if self.height < height_view:
                width_view = self.height * aspect
                height_view = self.height
            pos_x = round((self.width - width_view)/2)
            pos_y = round((self.height - height_view)/2)
            width_view = round(width_view)
            height_view = round(height_view)

        self.view_size = (width_view, height_view)
        self.view_pos  = [self.pos[0] + pos_x, self.pos[1] + pos_y]

    def decode_orientation(self):
        orientation = self.orientation
        if orientation == 'same':
            if Window.width > Window.height:
                orientation = 'landscape'
            else:
                orientation = 'portrait'
        elif orientation == 'opposite':
            if Window.width > Window.height:
                orientation = 'portrait'
            else:
                orientation = 'landscape'
        return orientation

    def screenshot_crop(self):
        pos_x = 0
        pos_y = 0
        if self.view_size[0] == round(self.width) and\
           self.view_size[1] != round(self.height):
            pos_y = (self.height - self.view_size[1])/2
        elif self.view_size[1] == round(self.height) and\
             self.view_size[0] != round(self.width):
            pos_x = (self.width - self.view_size[0])/2
        return (pos_x,pos_y, self.view_size[0],self.view_size[1])

    #############################################
    # File Utilities
    #############################################

    def capture_path(self,location, subdir ,name, ext):
        if platform == 'ios':
            storage = location.lower()
            if storage not in ['private', 'shared']:
                storage = 'shared'
            if storage == 'shared':
                return ''
            location = self._camera.get_app_documents_directory()  
        return join(self._default_subdir(location, subdir),
                    self._default_file_name(name, ext))
    
    def _default_subdir(self, location = '.', subdir=''):
        if not exists(location):
            location = '.'
        if not subdir:
            # Today's date
            subdir = datetime.now().strftime("%Y_%m_%d")
        path = join(location,subdir)
        if not exists(path):
            mkdir(path)
        return path
            
    def _default_file_name(self, name='', ext = '.jpg'):     
        if name:
            name = Path(name).stem
        else:
            name = datetime.now().strftime("%H_%M_%S_%f")[:-4]
        return name + ext            




__all__ = ('CameraOpenCV')

from kivy.logger import Logger
from kivy.clock import Clock
from kivy.graphics.texture import Texture
from kivy.utils import platform
from kivy.graphics import Color, Rectangle, Rotate, Fbo
import cv2    
from . import CameraBase

class CameraOpenCV(CameraBase):

    def __init__(self, **kwargs):
        self._device = None
        self._update_ev = None
        super(CameraOpenCV, self).__init__(**kwargs)

    def init_camera(self):
        self._format = 'bgr'
        if platform == 'win':
            self._index = self._index + cv2.CAP_DSHOW
        self._device = cv2.VideoCapture(self._index)
        self._device.set(cv2.CAP_PROP_FRAME_WIDTH,  self._resolution[0])
        self._device.set(cv2.CAP_PROP_FRAME_HEIGHT, self._resolution[1])
        ret, frame = self._device.read()
        self._resolution = (int(frame.shape[1]), int(frame.shape[0]))
        self.fps = self._device.get(cv2.CAP_PROP_FPS)
        if self.fps == 0 or self.fps == 1:
            self.fps = 1.0 / 30
        elif self.fps > 1:
            self.fps = 1.0 / self.fps
        self.crop = self._context.crop_for_aspect_orientation(*self._resolution)
        self.stopped = True

    def update(self, dt):
        if self.stopped:
            return
        if self._texture is None:
            self._texture = Texture.create(self._resolution)
            self._texture.flip_vertical()
            self._context.on_load()
        try:
            ret, frame = self._device.read()
            if ret:
                self._buffer = frame.reshape(-1)
                self._copy_to_gpu()
                if self.photo_capture:
                    self.photo_capture = False
                    cropped = frame[self.crop[1]: self.crop[1]+self.crop[3],
                                    self.crop[0]: self.crop[0]+self.crop[2], :]
                    cv2.imwrite(self.photo_path, cropped)
                    if self.photo_callback:
                        self.photo_callback(self.photo_path)
                if self.video_capture:
                    cropped = frame[self.crop[1]: self.crop[1]+self.crop[3],
                                    self.crop[0]: self.crop[0]+self.crop[2], :]
                    self.video_stream.write(cropped) 

        except Exception as e:
            Logger.exception('OpenCV: Couldn\'t get image from Camera')

    def start(self):
        self.stopped = False
        self.photo_capture = False
        self.video_capture = False
        if self._update_ev is not None:
            self._update_ev.cancel()
        self._update_ev = Clock.schedule_interval(self.update, 1/30) 

    def stop(self):
        self.stopped = True
        self._device = None
        if self._update_ev is not None:
            self._update_ev.cancel()
            self._update_ev = None

    def photo(self, path, callback):
        self.photo_capture = True
        self.photo_path = path
        self.photo_callback = callback

    def video_start(self, path, callback):
        self.video_capture = True
        self.video_path = path
        self.video_callback = callback
        size = (self.crop[2], self.crop[3])
        rate = Clock.get_fps()
        fourcc = cv2.VideoWriter_fourcc(*'mp4v')
        self.video_stream = cv2.VideoWriter(path, fourcc, rate, size)
        
    def video_stop(self):
        self.video_capture = False
        self.video_stream.release()

